/* This file was generated by scm2c from source file "handle.scm" */

extern int clever_cache_enable_P;

SCM_EXPORT int han_error_P P((HAND *han,char *caller));

SCM_EXPORT int key_error_P P((unsigned char *key_str,int k_len,int xcpt,char *caller));

SCM_EXPORT int bt_get P((HAND *han,unsigned char *key_str,int k_len,unsigned char *ans_str));

SCM_EXPORT int bt_next P((HAND *han,unsigned char *key_str,int k_len,unsigned char *ans_str));

SCM_EXPORT int bt_prev P((HAND *han,unsigned char *key_str,int k_len,unsigned char *ans_str));

SCM_EXPORT int bt_rem P((HAND *han,unsigned char *key_str,int k_len,unsigned char *ans_str));

SCM_EXPORT int bt_rem_range P((HAND *han,unsigned char *key_str,int k_len,unsigned char *key2_str,int k2_len));

SCM_EXPORT int bt_put P((HAND *han,unsigned char *key_str,int k_len,unsigned char *val_str,int v_len));

SCM_EXPORT int bt_write P((HAND *han,unsigned char *key_str,int k_len,unsigned char *val_str,int v_len));
