/* This file was generated by scm2c from source file "prev.scm" */

SCM_EXPORT ENTRY *chain_prev_find P((ENTRY *ent,int accmode,unsigned char *key_str,int k_len,int *pkt,ENTRY *prev_ent,int prev_pos));

SCM_EXPORT int str_gtr_P P((unsigned char *a_str,int a_pos,int a_len,unsigned char *b_str,int b_pos,int b_len));

SCM_EXPORT ENTRY *prev_prev_k_ent P((ENTRY *ent,unsigned char *key_str,int k_len,int level,int *pkt));

SCM_EXPORT ENTRY *prev_key_ent P((ENTRY *ent,unsigned char *key_str,int k_len,int level,int *pkt));

SCM_EXPORT ENTRY *chain_to_prev_ent P((ENTRY *from_ent,long goal_blk_num,unsigned char *goal_key_str,int key_len));

SCM_EXPORT int at_root_level_P P((SEGD *seg,unsigned char *blk));

SCM_EXPORT ENTRY *prev_blk_ent P((ENTRY *ent,int level));

SCM_EXPORT ENTRY *prev_find_prev_ent P((ENTRY *ent,int desired_level,int last_level,unsigned char *key_str,int k_len));
