/* This file was generated by scm2c from source file "scan.scm" */

SCM_EXPORT int bt_scan P((HAND *han,int operation,unsigned char *kstr1,int len1,unsigned char *kstr2,int len2,int_function func,long *long_tab,int *respkt,int blk_limit));

SCM_EXPORT int increment_string P((unsigned char *str,int len,int maxlen));

SCM_EXPORT int chain_scan P((ENTRY *ent,int operation,int *pkt,int *opkt,unsigned char *key_str,int_function func,long *long_tab,unsigned char *vstr,int *respkt,int wcb));

SCM_EXPORT int scan_loop P((unsigned char *blk,int operation,int *pkt,int *opkt,int_function func,long *long_tab,int *respkt,unsigned char *ckstr,int clen,unsigned char *vstr,int blksize));
