/* This file was generated by scm2c from source file "segs.scm" */

extern int seg_cntr;

extern SEGD *seg_chain;

extern LCK *seg_chain_lck;
SCM_EXPORT SEGD *new_seg P((unsigned char *filename,char *caller));

SCM_EXPORT void free_seg P((SEGD *seg));

SCM_EXPORT int check_file_P P((unsigned char *filename));

SCM_EXPORT SEGD *open_seg P((unsigned char *filename,int mutable_P));

SCM_EXPORT SEGD *open_segd P((unsigned char *filename,int mutable_P,int even_if_dirty_P));

SCM_EXPORT int close_seg P((SEGD *seg,int hammer_P));

extern unsigned char db_version_str[];

extern unsigned char db_authors_str[];

SCM_EXPORT SEGD *make_seg P((unsigned char *filename,int bsiz));

SCM_EXPORT int bt_open P((SEGD *seg,long blk_num,HAND *han,int wcb));

SCM_EXPORT void bt_open_init_han P((HAND *han,ENTRY *ent,int wcb));

SCM_EXPORT int bt_open_new P((SEGD *seg,long blk_num,HAND *han,int wcb,int typ));

SCM_EXPORT int bt_create P((SEGD *seg,int typ,HAND *han,int wcb));

SCM_EXPORT int bt_close P((HAND *han));
